/**
  *  \file server/objout.cc
  *  \brief Server: Object Output
  *
  *  This file contains classes to produce output in JSON format for
  *  various object requests. Most requests are directly mapped to
  *  script names.
  */

#include <memory>
#include <typeinfo>
#include "server/objout.h"
#include "cpluslib/format.h"
#include "server/jsonout.h"
#include "game/spec.h"
#include "game/global.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/team.h"
#include "game/racename.h"
#include "int/if/shipif.h"
#include "util/console.h"
#include "int/if/planetif.h"
#include "int/if/globalif.h"
#include "int/if/playerif.h"
#include "int/if/hullif.h"
#include "int/if/specif.h"
#include "int/if/ionif.h"
#include "int/if/mineif.h"
#include "int/if/ufoif.h"


/** Write value from script to JSON output. Automatically picks
    integer or string output depending on property type. When the
    property is not defined, it is not written to JSON output.
    \param w JSON writer
    \param con Object context
    \param scriptName Name of property to obtain from context
    \param jsonName Name of property in JSON output */
static void
writeValue(JsonWriter& w, IntContext& con, const char* scriptName, const char* jsonName)
{
    /* Look up name. This should not fail. */
    int32_t index = con.lookup(scriptName);
    if (index < 0) {
        console.write(LOG_ERROR, format("Unable to resolve name '%s'") << scriptName);
        return;
    }

    /* Get value */
    std::auto_ptr<IntValue> value(con.get(index));
    if (value.get()) {
        if (IntIntValue* iv = dynamic_cast<IntIntValue*>(value.get())) {
            // Is it a bird?
            w.writeInt(jsonName, iv->getValue());
        } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(value.get())) {
            // Is it a plane?
            w.writeString(jsonName, sv->getValue());
        } else {
            // What?
            console.write(LOG_ERROR, format("Unexpected type '%s'") << scriptName);
        }
    }
}



ServerObjectWriter::~ServerObjectWriter()
{ }

/************************** ServerSingularWriter *************************/

bool
ServerSingularWriter::isSame(ServerObjectWriter* other)
{
    return typeid(*this) == typeid(*other);
}

/*************************** ServerShipxyWriter **************************/

void
ServerShipxyWriter::write(JsonWriter& w)
{
    w.startArray("shipxy");
    for (int i = 0; i <= NUM_SHIPS; ++i) {
        if (!global_anyship_type.isExistant(i)) {
            w.writeNull();
        } else {
            IntShipContext c(i);
            w.startObject();
            writeValue(w, c, "LOC.X", "X");
            writeValue(w, c, "LOC.Y", "Y");
            writeValue(w, c, "MASS", "MASS");
            writeValue(w, c, "NAME", "NAME");
            writeValue(w, c, "OWNER$", "OWNER");
            writeValue(w, c, "PLAYED", "PLAYED");
            w.endObject();
        }
    }
    w.endArray();
}

/**************************** ServerShipWriter ***************************/

ServerShipWriter::ServerShipWriter(int sid)
    : sid(sid)
{ }

void
ServerShipWriter::write(JsonWriter& w)
{
    IntShipContext c(sid);
    w.startObject(format("ship%d", sid));
    writeValue(w, c, "AUX$", "AUX");
    writeValue(w, c, "AUX.AMMO", "AUX.AMMO");
    writeValue(w, c, "AUX.COUNT", "AUX.COUNT");
    writeValue(w, c, "BEAM$", "BEAM");
    writeValue(w, c, "BEAM.COUNT", "BEAM.COUNT");
    w.startObject("CARGO");
    writeValue(w, c, "CARGO.COLONISTS", "COLONISTS");
    writeValue(w, c, "CARGO.D", "D");
    writeValue(w, c, "CARGO.M", "M");
    writeValue(w, c, "CARGO.MONEY", "MC");
    writeValue(w, c, "CARGO.N", "N");
    writeValue(w, c, "CARGO.SUPPLIES", "SUPPLIES");
    writeValue(w, c, "CARGO.T", "T");
    w.endObject();
    writeValue(w, c, "COMMENT", "COMMENT");
    writeValue(w, c, "CREW", "CREW");
    writeValue(w, c, "DAMAGE", "DAMAGE");
    writeValue(w, c, "ENEMY$", "ENEMY");
    writeValue(w, c, "ENGINE$", "ENGINE");
    writeValue(w, c, "FCODE", "FCODE");
    writeValue(w, c, "HEADING$", "HEADING");
    writeValue(w, c, "HULL$", "HULL");
    writeValue(w, c, "LEVEL", "LEVEL");
    writeValue(w, c, "MISSION$", "MISSION");
    writeValue(w, c, "MISSION.INTERCEPT", "MISSION.INTERCEPT");
    writeValue(w, c, "MISSION.TOW", "MISSION.TOW");
    writeValue(w, c, "MOVE.ETA", "MOVE.ETA");
    writeValue(w, c, "MOVE.FUEL", "MOVE.FUEL");
    writeValue(w, c, "OWNER.REAL", "OWNER.REAL");
    writeValue(w, c, "SPEED$", "SPEED");
    writeValue(w, c, "WAYPOINT.DX", "WAYPOINT.DX");
    writeValue(w, c, "WAYPOINT.DY", "WAYPOINT.DY");
    if (getDisplayedTurn().getCurrentUniverse().getShip(sid).isTransporterActive(GShip::t_Transfer)) {
        w.startObject("TRANSFER");
        writeValue(w, c, "TRANSFER.SHIP.COLONISTS", "COLONISTS");
        writeValue(w, c, "TRANSFER.SHIP.D", "D");
        writeValue(w, c, "TRANSFER.SHIP.ID", "ID");
        writeValue(w, c, "TRANSFER.SHIP.M", "M");
        writeValue(w, c, "TRANSFER.SHIP.N", "N");
        writeValue(w, c, "TRANSFER.SHIP.SUPPLIES", "SUPPLIES");
        writeValue(w, c, "TRANSFER.SHIP.T", "T");
        w.endObject();
    }
    if (getDisplayedTurn().getCurrentUniverse().getShip(sid).isTransporterActive(GShip::t_Unload)) {
        w.startObject("UNLOAD");
        writeValue(w, c, "TRANSFER.UNLOAD.COLONISTS", "COLONISTS");
        writeValue(w, c, "TRANSFER.UNLOAD.D", "D");
        writeValue(w, c, "TRANSFER.UNLOAD.ID", "ID");
        writeValue(w, c, "TRANSFER.UNLOAD.M", "M");
        writeValue(w, c, "TRANSFER.UNLOAD.N", "N");
        writeValue(w, c, "TRANSFER.UNLOAD.SUPPLIES", "SUPPLIES");
        writeValue(w, c, "TRANSFER.UNLOAD.T", "T");
        w.endObject();
    }
    w.endObject();
}

bool
ServerShipWriter::isSame(ServerObjectWriter* other)
{
    ServerShipWriter* p = dynamic_cast<ServerShipWriter*>(other);
    return p != 0 && p->sid == sid;
}

/************************** ServerPlanetxyWriter *************************/

void
ServerPlanetxyWriter::write(JsonWriter& w)
{
    w.startArray("planetxy");
    for (int i = 0; i <= NUM_PLANETS; ++i) {
        if (!global_anyplanet_type.isExistant(i)) {
            w.writeNull();
        } else {
            IntPlanetContext c(i);
            w.startObject();
            writeValue(w, c, "BASE.YESNO", "BASE");
            writeValue(w, c, "LOC.X", "X");
            writeValue(w, c, "LOC.Y", "Y");
            writeValue(w, c, "NAME", "NAME");
            writeValue(w, c, "OWNER$", "OWNER");
            writeValue(w, c, "PLAYED", "PLAYED");
            w.endObject();
        }
    }
    w.endArray();
}

/*************************** ServerPlanetWriter **************************/

ServerPlanetWriter::ServerPlanetWriter(int pid)
    : pid(pid)
{ }

void
ServerPlanetWriter::write(JsonWriter& w)
{
    IntPlanetContext c(pid);
    w.startObject(format("planet%d", pid));
    writeValue(w, c, "BASE.BUILDING", "BASE.BUILDING");
    writeValue(w, c, "COLONISTS.HAPPY$", "COLONISTS.HAPPY");
    writeValue(w, c, "COLONISTS.SUPPORTED", "COLONISTS.SUPPORTED");
    writeValue(w, c, "COLONISTS.TAX", "COLONISTS.TAX");
    writeValue(w, c, "COMMENT", "COMMENT");
    writeValue(w, c, "DAMAGE", "DAMAGE");
    writeValue(w, c, "DEFENSE", "DEFENSE");
    writeValue(w, c, "DEFENSE.BASE", "DEFENSE.BASE");
    writeValue(w, c, "DEFENSE.BASE.WANT", "DEFENSE.BASE.WANT");
    writeValue(w, c, "DEFENSE.WANT", "DEFENSE.WANT");
    writeValue(w, c, "DENSITY.D", "DENSITY.D");
    writeValue(w, c, "DENSITY.M", "DENSITY.M");
    writeValue(w, c, "DENSITY.N", "DENSITY.N");
    writeValue(w, c, "DENSITY.T", "DENSITY.T");
    writeValue(w, c, "FACTORIES", "FACTORIES");
    writeValue(w, c, "FACTORIES.WANT", "FACTORIES.WANT");
    writeValue(w, c, "FCODE", "FCODE");
    writeValue(w, c, "FIGHTERS", "FIGHTERS");
    writeValue(w, c, "GROUND.D", "GROUND.D");
    writeValue(w, c, "GROUND.M", "GROUND.M");
    writeValue(w, c, "GROUND.N", "GROUND.N");
    writeValue(w, c, "GROUND.T", "GROUND.T");
    writeValue(w, c, "INDUSTRY$", "INDUSTRY");
    writeValue(w, c, "LEVEL", "LEVEL");
    w.startObject("G");
    writeValue(w, c, "COLONISTS", "COLONISTS");
    writeValue(w, c, "MINED.D", "D");
    writeValue(w, c, "MINED.M", "M");
    writeValue(w, c, "MINED.N", "N");
    writeValue(w, c, "MINED.T", "T");
    writeValue(w, c, "MONEY", "MC");
    writeValue(w, c, "SUPPLIES", "SUPPLIES");
    w.endObject();
    writeValue(w, c, "MINES", "MINES");
    writeValue(w, c, "MINES.WANT", "MINES.WANT");
    writeValue(w, c, "MISSION$", "MISSION");
    writeValue(w, c, "NATIVES", "NATIVES");
    writeValue(w, c, "NATIVES.GOV$", "NATIVES.GOV");
    writeValue(w, c, "NATIVES.HAPPY$", "NATIVES.HAPPY");
    writeValue(w, c, "NATIVES.RACE$", "NATIVES.RACE");
    writeValue(w, c, "NATIVES.TAX", "NATIVES.TAX");
    writeValue(w, c, "SHIPYARD.ACTION", "SHIPYARD.ACTION");
    writeValue(w, c, "SHIPYARD.ID", "SHIPYARD.ID");
    writeValue(w, c, "TECH.BEAM", "TECH.BEAM");
    writeValue(w, c, "TECH.ENGINE", "TECH.ENGINE");
    writeValue(w, c, "TECH.HULL", "TECH.HULL");
    writeValue(w, c, "TECH.TORPEDO", "TECH.TORPEDO");
    writeValue(w, c, "TEMP$", "TEMP");

    GPlanet& p = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
    if (p.hasBaseData() && p.getBaseBuildHull() != 0) {
        w.startObject("BUILD");
        writeValue(w, c, "BUILD.BEAM$", "BEAM");
        writeValue(w, c, "BUILD.BEAM.COUNT", "BEAM.COUNT");
        writeValue(w, c, "BUILD.ENGINE$", "ENGINE");
        writeValue(w, c, "BUILD.HULL$", "HULL");
        writeValue(w, c, "BUILD.QPOS", "QPOS");
        writeValue(w, c, "BUILD.TORP$", "TORP");
        writeValue(w, c, "BUILD.TORP.COUNT", "TORP.COUNT");
        w.endObject();
    }
    if (p.hasBaseData()) {
        w.startArray("STORAGE.AMMO");
        w.writeInt(0);
        for (int i = 1; i <= NUM_TORPS+1; ++i)  /* 11th element is fighters */
            w.writeInt(p.getBaseAmmoStore(i));
        w.endArray();
        w.startArray("STORAGE.BEAMS");
        w.writeInt(0);
        for (int i = 1; i <= NUM_BEAMS; ++i)
            w.writeInt(p.getBaseBeamStore(i));
        w.endArray();
        w.startArray("STORAGE.ENGINES");
        w.writeInt(0);
        for (int i = 1; i <= NUM_ENGINES; ++i)
            w.writeInt(p.getBaseEngineStore(i));
        w.endArray();
        w.startArray("STORAGE.HULLS");
        w.writeInt(0);
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
            w.writeInt(p.getBaseHullStoreSlot(i));
        w.endArray();
        w.startArray("STORAGE.LAUNCHERS");
        w.writeInt(0);
        for (int i = 1; i <= NUM_TORPS; ++i)
            w.writeInt(p.getBaseLauncherStore(i));
        w.endArray();
    }
    w.endObject();
}

bool
ServerPlanetWriter::isSame(ServerObjectWriter* other)
{
    ServerPlanetWriter* p = dynamic_cast<ServerPlanetWriter*>(other);
    return p != 0 && p->pid == pid;
}

/**************************** ServerMainWriter ***************************/

void
ServerMainWriter::write(JsonWriter& w)
{
    IntGlobalContext c;
    w.startObject("main");
    writeValue(w, c, "MY.INMSGS", "MY.INMSGS");
    writeValue(w, c, "MY.OUTMSGS", "MY.OUTMSGS");
    writeValue(w, c, "MY.RACE$", "MY.RACE");
    writeValue(w, c, "MY.RACE.ID", "MY.RACE.ID");
    writeValue(w, c, "MY.RACE.MISSION", "MY.RACE.MISSION");
    writeValue(w, c, "MY.VCRS", "MY.VCRS");
    writeValue(w, c, "SYSTEM.GAMETYPE$", "SYSTEM.GAMETYPE");
    writeValue(w, c, "SYSTEM.HOST", "SYSTEM.HOST");
    writeValue(w, c, "SYSTEM.HOST$", "SYSTEM.HOST$");
    writeValue(w, c, "SYSTEM.HOSTVERSION", "SYSTEM.HOSTVERSION");
    writeValue(w, c, "SYSTEM.REGSTR1", "SYSTEM.REGSTR1");
    writeValue(w, c, "SYSTEM.REGSTR2", "SYSTEM.REGSTR2");
    writeValue(w, c, "SYSTEM.VERSION", "SYSTEM.VERSION");
    writeValue(w, c, "TURN", "TURN");
    writeValue(w, c, "TURN.DATE", "TURN.DATE");
    writeValue(w, c, "TURN.TIME", "TURN.TIME");
    w.endObject();
}

/*************************** ServerPlayerWriter **************************/

void
ServerPlayerWriter::write(JsonWriter& w)
{
    w.startArray("player");
    w.writeNull();
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        IntPlayerContext c(i);
        w.startObject();
        writeValue(w, c, "BASES", "BASES");
        writeValue(w, c, "PLANETS", "PLANETS");
        writeValue(w, c, "RACE", "RACE");
        writeValue(w, c, "RACE$", "RACE$");
        writeValue(w, c, "RACE.ADJ", "RACE.ADJ");
        writeValue(w, c, "RACE.ID", "RACE.ID");
        writeValue(w, c, "RACE.MISSION", "RACE.MISSION");
        writeValue(w, c, "RACE.SHORT", "RACE.SHORT");
        writeValue(w, c, "SCORE", "SCORE");
        writeValue(w, c, "SHIPS", "SHIPS");
        writeValue(w, c, "SHIPS.CAPITAL", "SHIPS.CAPITAL");
        writeValue(w, c, "SHIPS.FREIGHTERS", "SHIPS.FREIGHTERS");
        w.endObject();
    }
    w.endArray();
}

/**************************** ServerHullWriter ***************************/

ServerHullWriter::ServerHullWriter(int hid)
    : hid(hid)
{ }

void
ServerHullWriter::write(JsonWriter& w)
{
    IntHullContext c(hid);
    w.startObject(format("hull%d", hid));
    writeValue(w, c, "BEAM.MAX", "BEAM.MAX");
    writeValue(w, c, "CARGO.MAX", "CARGO.MAX");
    writeValue(w, c, "CARGO.MAXFUEL", "CARGO.MAXFUEL");
    w.startObject("COST");
    writeValue(w, c, "COST.D", "D");
    writeValue(w, c, "COST.M", "M");
    writeValue(w, c, "COST.MC", "MC");
    writeValue(w, c, "COST.T", "T");
    w.endObject();
    writeValue(w, c, "CREW.NORMAL", "CREW.NORMAL");
    writeValue(w, c, "ENGINE.COUNT", "ENGINE.COUNT");
    writeValue(w, c, "FIGHTER.BAYS", "FIGHTER.BAYS");
    writeValue(w, c, "IMAGE", "IMAGE");
    writeValue(w, c, "MASS", "MASS");
    writeValue(w, c, "NAME", "NAME");
    writeValue(w, c, "TECH", "TECH");
    writeValue(w, c, "TORP.LMAX", "TORP.LMAX");
    w.endObject();
}

bool
ServerHullWriter::isSame(ServerObjectWriter* other)
{
    ServerHullWriter* p = dynamic_cast<ServerHullWriter*>(other);
    return p != 0 && p->hid == hid;
}

/**************************** ServerTorpWriter ***************************/

void
ServerTorpWriter::write(JsonWriter& w)
{
    w.startArray("torp");
    w.writeNull();
    for (int i = 1; i <= NUM_TORPS; ++i) {
        IntTorpedoContext c(i);
        w.startObject();
        w.startObject("TORPCOST");
        writeValue(w, c, "COST.D", "D");
        writeValue(w, c, "COST.M", "M");
        writeValue(w, c, "COST.MC", "MC");
        writeValue(w, c, "COST.T", "T");
        w.endObject();
        writeValue(w, c, "DAMAGE", "DAMAGE");
        writeValue(w, c, "KILL", "KILL");
        writeValue(w, c, "NAME", "NAME");
        writeValue(w, c, "TECH", "TECH");

        IntLauncherContext lc(i);
        writeValue(w, lc, "MASS", "MASS");
        w.startObject("TUBECOST");
        writeValue(w, lc, "COST.D", "D");
        writeValue(w, lc, "COST.M", "M");
        writeValue(w, lc, "COST.MC", "MC");
        writeValue(w, lc, "COST.T", "T");
        w.endObject();
        w.endObject();
    }
    w.endArray();
}

/**************************** ServerBeamWriter ***************************/

void
ServerBeamWriter::write(JsonWriter& w)
{
    w.startArray("beam");
    w.writeNull();
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        IntBeamContext c(i);
        w.startObject();
        w.startObject("COST");
        writeValue(w, c, "COST.D", "D");
        writeValue(w, c, "COST.M", "M");
        writeValue(w, c, "COST.MC", "MC");
        writeValue(w, c, "COST.T", "T");
        w.endObject();
        writeValue(w, c, "DAMAGE", "DAMAGE");
        writeValue(w, c, "KILL", "KILL");
        writeValue(w, c, "MASS", "MASS");
        writeValue(w, c, "NAME", "NAME");
        writeValue(w, c, "TECH", "TECH");
        w.endObject();
    }
    w.endArray();
}

/*************************** ServerEngineWriter **************************/

void
ServerEngineWriter::write(JsonWriter& w)
{
    w.startArray("engine");
    w.writeNull();
    for (int i = 1; i <= NUM_ENGINES; ++i) {
        IntEngineContext c(i);
        w.startObject();
        w.startObject("COST");
        writeValue(w, c, "COST.D", "D");
        writeValue(w, c, "COST.M", "M");
        writeValue(w, c, "COST.MC", "MC");
        writeValue(w, c, "COST.T", "T");
        w.endObject();
        w.startArray("FUELFACTOR");
        for (int speed = 0; speed <= 9; ++speed)
            w.writeInt(getEngine(i).getFuelFactor(speed));
        w.endArray();
        writeValue(w, c, "NAME", "NAME");
        writeValue(w, c, "SPEED$", "SPEED");
        writeValue(w, c, "TECH", "TECH");
        w.endObject();
    }
    w.endArray();
}

/*************************** ServerStormWriter ***************************/

void
ServerStormWriter::write(JsonWriter& w)
{
    w.startArray("zstorm");
    for (int i = global_ion_type.getNextNoWrap(0, false); i != 0; i = global_ion_type.getNextNoWrap(i, false)) {
        IntIonContext c(i);
        w.startObject();
        writeValue(w, c, "HEADING$", "HEADING");
        writeValue(w, c, "ID", "ID");
        writeValue(w, c, "LOC.X", "X");
        writeValue(w, c, "LOC.Y", "Y");
        writeValue(w, c, "NAME", "NAME");
        writeValue(w, c, "RADIUS", "RADIUS");
        writeValue(w, c, "SPEED$", "SPEED");
        writeValue(w, c, "STATUS$", "STATUS");
        writeValue(w, c, "VOLTAGE", "VOLTAGE");
        w.endObject();
    }
    w.endArray();
}

/************************* ServerMinefieldWriter *************************/

void
ServerMinefieldWriter::write(JsonWriter& w)
{
    w.startArray("zmine");
    for (int i = global_minefield_type.getNextNoWrap(0, false); i != 0; i = global_minefield_type.getNextNoWrap(i, false)) {
        IntMinefieldContext c(i);
        w.startObject();
        writeValue(w, c, "ID", "ID");
        writeValue(w, c, "LASTSCAN", "LASTSCAN");
        writeValue(w, c, "LOC.X", "X");
        writeValue(w, c, "LOC.Y", "Y");
        writeValue(w, c, "OWNER$", "OWNER");
        writeValue(w, c, "RADIUS", "RADIUS");
        writeValue(w, c, "SCANNED", "SCANNED");
        writeValue(w, c, "TYPE$", "TYPE");
        writeValue(w, c, "UNITS", "UNITS");
        w.endObject();
    }
    w.endArray();
}

/**************************** ServerUfoWriter ****************************/

void
ServerUfoWriter::write(JsonWriter& w)
{
    w.startArray("zufo");
    GUfoContainer& uc = getDisplayedTurn().getCurrentUniverse().getUfos();
    for (GUfoContainer::iterator i = uc.begin(); i != uc.end(); ++i) {
        IntUfoContext c(uc, i);
        w.startObject();
        writeValue(w, c, "COLOR.EGA", "COLOR");
        writeValue(w, c, "HEADING$", "HEADING");
        writeValue(w, c, "ID", "ID");
        writeValue(w, c, "INFO1", "INFO1");
        writeValue(w, c, "INFO2", "INFO2");
        writeValue(w, c, "KEEP", "KEEP");
        writeValue(w, c, "LASTSCAN", "LASTSCAN");
        writeValue(w, c, "LOC.X", "X");
        writeValue(w, c, "LOC.Y", "Y");
        writeValue(w, c, "MOVE.DX", "MOVE.DX");
        writeValue(w, c, "MOVE.DY", "MOVE.DY");
        writeValue(w, c, "NAME", "NAME");
        writeValue(w, c, "RADIUS", "RADIUS");
        writeValue(w, c, "SPEED$", "SPEED");
        writeValue(w, c, "TYPE", "TYPE");
        writeValue(w, c, "VISIBLE.PLANET", "VISIBLE.PLANET");
        writeValue(w, c, "VISIBLE.SHIP", "VISIBLE.SHIP");
        w.endObject();
    }
    w.endArray();
}

/************************** ServerMessageWriter **************************/

ServerMessageWriter::ServerMessageWriter(int index)
    : index(index)
{ }

void
ServerMessageWriter::write(JsonWriter& w)
{
    GInbox& in = getDisplayedTurn().getInbox();
    w.startObject(format("msg%d", index));
    w.writeString("GROUP", in.getHeading(index-1));
    w.writeString("TEXT", in.getText(index-1));
    w.endObject();
}

bool
ServerMessageWriter::isSame(ServerObjectWriter* other)
{
    ServerMessageWriter* p = dynamic_cast<ServerMessageWriter*>(other);
    return p != 0 && p->index == index;
}

/************************** ServerTruehullWriter *************************/

void
ServerTruehullWriter::write(JsonWriter& w)
{
    w.startArray("truehull");
    w.writeNull();
    for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
        w.startArray();
        for (int idx = 1; idx <= NUM_HULLS_PER_PLAYER; ++idx)
            w.writeInt(getTruehull(pl, idx));
        w.endArray();
    }
    w.endArray();
}
